from .motordriver import driver 
from threading import Timer
import numpy as np


DRIVECONTROL_PERIOD = 500/1000

POSITION_TOLERANCE = 1

MAX_DUTY_CYCLE = 0.75
MIN_DUTY_CYCLE = 0.25

#angle value at which 
STATE_TRANSISTION_ANGLE = 5.0

#angle transition overlap - prevent state oscilations at border - not currently used
STATE_OVERLAP_ANGLE = 2.0

TANK_STATE_K = 0.1
TANK_STATE_B = 0.1 # not currently used

COMP_STATE_STEER_K = 0.1
COMP_STATE_STEER_B = 0.1 # not currently used

COMP_STATE_DRIVE_K = 0.1
COMP_STATE_DRIVE_B = 0.1 # not currently used

def saturation(duty_cycle) :
    if duty_cycle > MAX_DUTY_CYCLE :
        duty_cycle = MAX_DUTY_CYCLE
    elif duty_cycle < MIN_DUTY_CYCLE :
        duty_cycle = MIN_DUTY_CYCLE

    return duty_cycle

def distanceBetween(cord1, cord2) :
        delta_x = np.int16(cord1[0]) - cord2[0]
        delta_y = np.int16(cord1[1]) - cord2[1]

        return np.int16(np.sqrt([np.square([delta_x]) + np.square([delta_y])]))[0][0]

def angleBetween(cord1, cord2, orientation) :
    delta_x = np.int16(cord1[0]) - cord2[0]
    delta_y = np.int16(cord1[1]) - cord2[1]

    global_angle = np.int16(np.arctan2([delta_x], [delta_y])*180/np.pi)[0]

   #convert from -180-180 to 0-360
    if global_angle < 0 :
        global_angle += 360
   
    difference = global_angle - orientation

    short_angle = difference if np.absolute(difference) < 180 else difference - 360

    return short_angle

class drivecontrol :

    ###############"PRIVATE"###############

    position = np.array([np.uint8(0), np.uint8(0)])
    orientation = np.uint16(0)

    last_position = np.array([np.uint8(0), np.uint8(0)])
    last_orientation = np.uint16(0)

    destination = np.array([np.uint8(0), np.uint8(0)])

    distance = np.uint16(0)
    delta_orientation = np.int16(0)

    motorDriver = driver()


    def updateDelta(self) : #updated to 
        
        self.distance = distanceBetween(self.destination, self.position)

        self.delta_orientation = angleBetween(self.destination, self.position, self.orientation)

        if self.distance > POSITION_TOLERANCE :
            self.motorDriver.enableMotors()
        else :
            self.motorDriver.disableMotors()

    #ISR subfunctions
    def tankPWM(self) :
        self.motorDriver.writeLeftPWM(saturation(0.5 + TANK_STATE_K*self.delta_orientation))
        self.motorDriver.writeRightPWM(saturation(0.5 - TANK_STATE_K*self.delta_orientation))
        
    def compensationPWM(self) :
        left_pwm = saturation(0.5 + COMP_STATE_DRIVE_K*self.distance) + COMP_STATE_STEER_K*self.delta_orientation
        right_pwm = saturation(0.5 + COMP_STATE_DRIVE_K*self.distance) - COMP_STATE_STEER_K*self.delta_orientation

        self.motorDriver.writeLeftPWM(left_pwm)
        self.motorDriver.writeRightPWM(right_pwm)
    
    #ISR called for each interrupt
    def pwmISR(self) :
        match np.absolute(self.delta_orientation) < STATE_TRANSISTION_ANGLE:
            case 1:
                self.compensationPWM()
            case 0:
                self.tankPWM()
        self.ISR = Timer(DRIVECONTROL_PERIOD, self.pwmISR).start()

    #member ISR timer
    ISR = Timer(DRIVECONTROL_PERIOD, pwmISR)

    ###############"PUBLIC"###############

    def setupDrive(self) :
        self.motorDriver.setupMotors()
    
    def enableDrive(self) :
        self.ISR.start()

    def disableDrive(self) :
        self.ISR.cancel()

    #argument type of array size 2
    def updateTarget(self, XY) :
        self.destination = np.copy(XY)
        self.updateDelta()

    def updatePosition(self, XY, orientation) :
        self.position = np.copy(XY)
        self.orientation = orientation
        self.updateDelta()
        